;**************************************************************************
;	Filename: 	BL33F.EXP16.S
;	Project:	BL24H33F
;
;   Copyright (c) 2006 - Microsockets - All rights reserved.
;   A division of Sanctuary Software Studio, Inc.
;	www.microsockets.com
;
;	Descripton:
;	Assembly boot loader for Microchip PIC 24H and 33F processors 
;	supporting serial port based in circuit programming
;
; 	The bootloader is placed at the top of program space in the last
;   erasable block (Top of memory - 0x400).  The reset vector must be
;	configured to enter the bootloader in the application program. In
;   addition, the application program must have a GOTO ENTRYPOINT at
;	address 0x000200 to support entering the application from the 
;	bootloader.  The placement of the bootloader and the application
;   connection are configured through the linker (GLD) files.
;
;   Explorer 16 Board Version
; 	Configured for operation on the Microchip Explorer 16 evaluation board
;	o UART 2 for communications
; 	o 8MHz onboard oscillator - 40Mhz FCYC
;
		.include	"p33Fxxxx.inc"
;**************************************************************************
;	Device Configuration

		config __FGS, GSS_OFF & GCP_OFF & GWRP_OFF
		config __FOSCSEL, FNOSC_PRIPLL & IESO_OFF 
		config __FOSC, FCKSM_CSDCMD & OSCIOFNC_OFF & POSCMD_XT
		config __FWDT, FWDTEN_OFF & WINDIS_OFF & WDTPRE_PR128 & WDTPOST_PS32768
		config __FPOR, FPWRT_PWR128

;**************************************************************************
;	Global Definitions
		.equ VERSION, 0x01
		.equ FCYC, 40000000
		.equ BAUD, 38400
		.equ BAUDGEN, ((FCYC/BAUD/16)-1)
		.equ BOOTDELAY,5
		.equ T23LOAD, ((0xFFFFFFFF)-((FCYC/64)*BOOTDELAY))
		.equ T2LOAD,((T23LOAD) & (0x0000FFFF))
		.equ T3LOAD, ((T23LOAD) >> 16)
		.equ REQBLCHAR, 0x41
		.equ RWNUMOP, 64
		.equ APPENTRY, 0x000200

;**************************************************************************
;	Global Declarations
		.global		__reset

;**************************************************************************
;	Data Declarations
		.bss 
RWBUF:	.space (3*RWNUMOP)

;**************************************************************************
;	Main Program
		.text
__reset:
		; initialize the oscillator
		MOV #0x0300, W0
		MOV W0,OSCCON
		MOV	#0x0000, W0
		MOV W0,CLKDIV
		MOV #38, W0
		MOV	W0, PLLFBD 
		
		; initialize the stack
		MOV	#__SP_init, W15			; initialize stack pointer
		MOV #__SPLIM_init, W0		; initialize the stack limit
		MOV W0, SPLIM
		NOP
		; initialize the UART for configured baud, 8N1
		BCLR TRISF, #5 				; TX as output
		BSET TRISF, #13				; RX as input
		MOV #BAUDGEN, W0			; Configure baud rate
		MOV W0, U2BRG
		MOV #0x8000, W0				; Fixed baud, 8N1
		MOV W0, U2MODE
		MOV #0x8400, W0				; Enable transmitter
		MOV W0, U2STA

		; initialize the start up timer
		MOV	#0x0028, W0				; stop, 32-bit, /64, FCYC
		MOV	W0, T2CON
		MOV #T3LOAD, W1				; preload most sig word
		MOV #T2LOAD, W0				; preload least sig word
		MOV W1, TMR3HLD
		MOV W0, TMR2
		BCLR IFS0, #T3IF			; clear the interrupt flag
		BSET T2CON, #15				; start the timer
		MOV #REQBLCHAR, W1			; load the target character for bootloader
waitforactivate:
		BTST IFS1, #U2RXIF			; check for received character
		BRA Z, checkstarttimer 		; no char received, check timer
		MOV	U2RXREG, W0				; get the character
		BCLR IFS1, #U2RXIF			; clear character received
		CP	W0, W1					; is it our character?
		BRA Z, ackuartbootload		; match - go to uart bl mode
checkstarttimer:
		BTSS IFS0, #T3IF			; check for timer expiration
		BRA waitforactivate			; wait for timer expiration
bootapplication:
		GOTO APPENTRY				; jump to entry point
		BRA	bootapplication			; TODO: boot the application firmware
ackuartbootload:
		MOV #'*', W3				; acknowledge bootloader mode
		CALL UART_TransmitChar
uartbootload:
		CLR	W7						; clear the checksum
		CALL UART_ReceiveChar		; get a command byte
		MOV.B W4, W8				; W8 holds the command
		MOV #'V',W5					; check for version command
		CP.B W8, W5					
		BRA Z, uartrecvchecksum
		MOV #'B', W5				; check for boot command
		CP.B W8, W5					
		BRA Z, uartrecvchecksum
		CLR W6						; receive the target address (3bytes)
		CALL UART_ReceiveChar		; build dword address in w6, w5
		MOV.B W4, W6
		CALL UART_ReceiveChar
		SL W4, #8, W5
		CALL UART_ReceiveChar
		MOV.B W4, W5		
		
		MOV.B #'W', W4				; check for write command
		CP.B W4, W8
		BRA NZ, uartrecvchecksum
		
		MOV #RWNUMOP, W0			; receive x operands
		MOV #RWBUF, W1				; point to buffer
uartbufferwritedata:
		CALL UART_ReceiveChar		; receive 3-bytes per instruction
		MOV.B W4, [W1++]			; low word, low byte
		CALL UART_ReceiveChar
		MOV.B W4, [W1++]			; low word, high byte
		CALL UART_ReceiveChar
		MOV.B W4, [W1++]			; high word, low byte
		DEC W0, W0
		BRA NZ, uartbufferwritedata
		
uartrecvchecksum:
		CALL UART_ReceiveChar		; receive checskum character
		SUB.B W7, W4, W7			; subtract from calculated, 0=good
		SUB.B W7, W4, W7			; subtract from calculated, 0=good
		BRA NZ, uartnack			; reject command if non-zero checksum
uartdispatch:
		MOV	#'V', W3
		CP.B W3, W8
		BRA Z, uartversion			; dispatch version command
		MOV #'B', W3
		CP.B W3, W8
		BRA Z, uartboot				; dispatch boot command
		MOV #'W', W3
		CP.B W3, W8
		BRA Z, uartwrite			; dispatch write command
		MOV #'E', W3
		CP.B W3, W8
		BRA Z, uarterase			; dispatch erase command
		MOV #'R', W3
		CP.B W3, W8
		BRA Z, uartread				; dispatch read command
		BRA uartnack

;**************************************************************************
;	uartversion - process the version command
uartversion:
		CLR	W7						; clear checksum
		CALL UART_TransmitChar
		MOV.B #VERSION, W3
		CALL UART_TransmitChar		; send the version
		MOV.B W7, W3
		CALL UART_TransmitChar		; send the checksum
		BRA uartbootload

;**************************************************************************
;	uarterase - process the erase command
uarterase:
		CLR W7						; clear the checksum
		CLR	W0						; clear the return code
		CALL FLASH_ErasePage
		MOV	#'E', W3
		CALL UART_TransmitChar		; send the command
		MOV.B W0, W3				; 
		CALL UART_TransmitChar		; send the return code
		MOV.B W7, W3
		CALL UART_TransmitChar		; send the checksum
		BRA uartbootload

;**************************************************************************
;	uartwrite - process the write command
uartwrite:
		CLR W7						; clear the checksum
		CLR	W0						; clear the return code
		CALL FLASH_WritePage
		MOV	#'W', W3
		CALL UART_TransmitChar		; send the command
		MOV.B W0, W3				; 
		CALL UART_TransmitChar		; send the return code
		MOV.B W7, W3
		CALL UART_TransmitChar		; send the checksum
		BRA uartbootload

;**************************************************************************
;	uartread - process the read command
uartread:
		CLR W7						; clear the checksum
		CLR	W0						; clear the return code
		CALL FLASH_ReadPage
		MOV	#'R', W3
		CALL UART_TransmitChar		; send the command
		MOV #RWNUMOP, W0			; send x operands
		MOV #RWBUF, W1				; point to buffer
uartbufferreaddata:
		MOV.B [W1++], W3			; low word, low byte
		CALL UART_TransmitChar
		MOV.B [W1++], W3			; low word, high byte
		CALL UART_TransmitChar
		MOV.B [W1++], W3			; high word, low byte
		CALL UART_TransmitChar
		DEC W0, W0
		BRA NZ, uartbufferreaddata		
		MOV.B W7, W3
		CALL UART_TransmitChar		; send the checksum
		BRA uartbootload
		
;**************************************************************************
;	uartboot - process the boot command
uartboot:
		CLR W7						; clear the checksum
		MOV	#'B', W3
		CALL UART_TransmitChar		; send the command
		MOV.B W7, W3
		CALL UART_TransmitChar		; send the checksum
uartbootxmitwait:
		BTSS IFS1, #U2TXIF
		BRA	uartbootxmitwait		; check for command gone
		BRA bootapplication			; start the application

uartnack:
		MOV	#'!', W3
		CALL UART_TransmitChar
		BRA	uartbootload

;**************************************************************************
;	UART_ReceiveChar
;	Wait for a character to come in on the UART
;   Returns W4 - character received
UART_ReceiveChar:
		BTSS IFS1, #U2RXIF
		BRA	UART_ReceiveChar
		MOV U2RXREG, W4
		BCLR IFS1, #U2RXIF
		ADD.B W4, W7, W7				; update the checksum
		RETURN

;**************************************************************************
;	UART_TransmitChar
;	Sends the character in W3 out the serial port, waits for the transmit
;	buffer to have space and then sends
;   Returns: Nothing
UART_TransmitChar:
		BTSS IFS1, #U2TXIF
		BRA	UART_TransmitChar
		MOV W3, U2TXREG
		BCLR IFS1, #U2TXIF
		ADD.B W3, W7, W7				; update the checksum
		RETURN

;**************************************************************************
;	FLASH_ErasePage
;	Erase the block at W6:W5
;   Returns: W0 contains error code (0 for success)
FLASH_ErasePage:
		MOV	#0x4042, W0
		MOV W0, NVMCON					; initialize NVM control
		MOV	W6, TBLPAG					; set the table page
		TBLWTL W5, [W5]					; set the table offset
		DISI #5							; no interrupts for 5 instructions
		MOV	#0x55, W0					; key access the flash
		MOV W0, NVMKEY
		MOV #0xAA, W0
		MOV W0, NVMKEY
		BSET NVMCON, #WR				; perform the erase
		NOP
		NOP
		CLR W0							; no errors
		RETURN

;**************************************************************************
;	FLASH_WritePage
;	Write the block at W6:W5
;   Returns: W0 contains error code (0 for success)
FLASH_WritePage:
		MOV #0X4001, W0
		MOV W0, NVMCON					; initialize NVM control
		MOV	W6, TBLPAG					; set page boundary
		MOV #RWBUF, W8
		MOV #RWNUMOP, W1
FLASH_WritePage_Copy:
		MOV.B [W8++], W2
		SL W2, #8, W2
		MOV.B [W8++], W2
		SWAP W2
		CLR W3
		MOV.B [W8++], W3
		TBLWTL W2, [W5]
		TBLWTH W3, [W5++]
		DEC W1, W1
		BRA NZ, FLASH_WritePage_Copy
 		DISI #5							; no interrupts for 5 instructions
		MOV	#0x55, W0					; key access the flash
		MOV W0, NVMKEY
		MOV #0xAA, W0
		MOV W0, NVMKEY
		BSET NVMCON, #WR				; perform the erase
		NOP
		NOP
		CLR W0							; no errors
		RETURN

;**************************************************************************
;	FLASH_ReadPage
;	Read the block at W6:W5
;   Returns: W0 contains error code (0 for success)
FLASH_ReadPage:
		MOV W6, TBLPAG					; set the table paging
		MOV #RWBUF, W8					; point to r/w buffer
		MOV #RWNUMOP, W1				; loop for # of operands
FLASH_ReadPage_Copy:
		TBLRDL [W5], W2					; read the operands
		TBLRDH [W5++], W3
		MOV.B W2,[W8++]					; low word, low byte
		SWAP W2
		MOV.B W2, [W8++]				; low word, high byte
		MOV.B W3, [W8++]				; high word, low byte
		DEC W1, W1						; 3 bytes per instruction
		BRA NZ, FLASH_ReadPage_Copy		
		CLR	W0							; no errors
		RETURN

;**************************************************************************
;	End of Program Code
.end									
